@@ -1,23 +1,19 @@
-/* $OpenLDAP: pkg/ldap/libraries/libldap/compare.c,v 1.9.6.5 2002/01/04 20:38:19 kurt Exp $ */
-/*
- * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
- * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
- */
-/*  Portions
- *  Copyright (c) 1990 Regents of the University of Michigan.
- *  All rights reserved.
+/* $OpenLDAP: pkg/ldap/libraries/libldap/compare.c,v 1.29.2.2 2007/08/31 23:13:55 quanah Exp $ */
+/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- *  compare.c
+ * Copyright 1998-2007 The OpenLDAP Foundation.
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted only as authorized by the OpenLDAP
+ * Public License.
+ *
+ * A copy of this license is available in the file LICENSE in the
+ * top-level directory of the distribution or, alternatively, at
+ * <http://www.OpenLDAP.org/license.html>.
  */
-
-/* The compare request looks like this:
- *	CompareRequest ::= SEQUENCE {
- *		entry	DistinguishedName,
- *		ava	SEQUENCE {
- *			type	AttributeType,
- *			value	AttributeValue
- *		}
- *	}
+/* Portions Copyright (c) 1990 Regents of the University of Michigan.
+ * All rights reserved.
  */
 
 #include "portable.h"
@@ -29,6 +25,17 @@
 #include <ac/time.h>
 
 #include "ldap-int.h"
+#include "ldap_log.h"
+
+/* The compare request looks like this:
+ *	CompareRequest ::= SEQUENCE {
+ *		entry	DistinguishedName,
+ *		ava	SEQUENCE {
+ *			type	AttributeType,
+ *			value	AttributeValue
+ *		}
+ *	}
+ */
 
 /*
  * ldap_compare_ext - perform an ldap extended compare operation.  The dn
@@ -36,7 +43,7 @@
  * attr and value) are supplied.  The msgid of the response is returned.
  *
  * Example:
- *	struct berval bvalue = { "secret", strlen("secret") };
+ *	struct berval bvalue = { "secret", sizeof("secret")-1 };
  *	rc = ldap_compare( ld, "c=us@cn=bob",
  *		"userPassword", &bvalue,
  *		sctrl, cctrl, &msgid )
@@ -53,6 +60,7 @@ ldap_compare_ext(
 {
 	int rc;
 	BerElement	*ber;
+	ber_int_t	id;
 
 	Debug( LDAP_DEBUG_TRACE, "ldap_compare\n", 0, 0, 0 );
 
@@ -71,9 +79,11 @@ ldap_compare_ext(
 		return( LDAP_NO_MEMORY );
 	}
 
-	if ( ber_printf( ber, "{it{s{sON}N}", /* '}' */
-		++ld->ld_msgid,
-		LDAP_REQ_COMPARE, dn, attr, bvalue ) == -1 )
+	LDAP_NEXT_MSGID(ld, id);
+	rc = ber_printf( ber, "{it{s{sON}N}", /* '}' */
+		id,
+		LDAP_REQ_COMPARE, dn, attr, bvalue );
+	if ( rc == -1 )
 	{
 		ld->ld_errno = LDAP_ENCODING_ERROR;
 		ber_free( ber, 1 );
@@ -92,20 +102,9 @@ ldap_compare_ext(
 		return( ld->ld_errno );
 	}
 
-#ifndef LDAP_NOCACHE
-	if ( ld->ld_cache != NULL ) {
-		if ( ldap_check_cache( ld, LDAP_REQ_COMPARE, ber ) == 0 ) {
-			ber_free( ber, 1 );
-			ld->ld_errno = LDAP_SUCCESS;
-			*msgidp = ld->ld_msgid;
-			return( ld->ld_errno );
-		}
-		ldap_add_request_to_cache( ld, LDAP_REQ_COMPARE, ber );
-	}
-#endif /* LDAP_NOCACHE */
 
 	/* send the message */
-	*msgidp = ldap_send_initial_request( ld, LDAP_REQ_COMPARE, dn, ber );
+	*msgidp = ldap_send_initial_request( ld, LDAP_REQ_COMPARE, dn, ber, id );
 	return ( *msgidp < 0 ? ld->ld_errno : LDAP_SUCCESS );
 }
 
@@ -127,6 +126,8 @@ ldap_compare(
 	int msgid;
 	struct berval bvalue;
 
+	assert( value != NULL );
+
 	bvalue.bv_val = (char *) value;
 	bvalue.bv_len = (value == NULL) ? 0 : strlen( value );
 
@@ -152,7 +153,7 @@ ldap_compare_ext_s(
 	if (  rc != LDAP_SUCCESS )
 		return( rc );
 
-	if ( ldap_result( ld, msgid, 1, (struct timeval *) NULL, &res ) == -1 )
+	if ( ldap_result( ld, msgid, LDAP_MSG_ALL, (struct timeval *) NULL, &res ) == -1 || !res )
 		return( ld->ld_errno );
 
 	return( ldap_result2error( ld, res, 1 ) );
@@ -167,6 +168,8 @@ ldap_compare_s(
 {
 	struct berval bvalue;
 
+	assert( value != NULL );
+
 	bvalue.bv_val = (char *) value;
 	bvalue.bv_len = (value == NULL) ? 0 : strlen( value );
 
