@@ -1,18 +1,21 @@
-/* $OpenLDAP: pkg/ldap/libraries/libldap/modrdn.c,v 1.12.6.5 2002/01/04 20:38:20 kurt Exp $ */
-/*
- * Copyright 1998-2002 The OpenLDAP Foundation, All Rights Reserved.
- * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
- */
-/*  Portions
- *  Copyright (c) 1990 Regents of the University of Michigan.
- *  All rights reserved.
+/* $OpenLDAP: pkg/ldap/libraries/libldap/modrdn.c,v 1.30.2.2 2007/08/31 23:13:55 quanah Exp $ */
+/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- *  modrdn.c
- */
-/*
- * Support for MODIFYDN REQUEST V3 (newSuperior) by:
+ * Copyright 1998-2007 The OpenLDAP Foundation.
+ * All rights reserved.
  *
- * Copyright 1999, Juan C. Gomez, All rights reserved.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted only as authorized by the OpenLDAP
+ * Public License.
+ *
+ * A copy of this license is available in the file LICENSE in the
+ * top-level directory of the distribution or, alternatively, at
+ * <http://www.OpenLDAP.org/license.html>.
+ */
+/* Portions Copyright (c) 1990 Regents of the University of Michigan.
+ * All rights reserved.
+ */
+/* Copyright 1999, Juan C. Gomez, All rights reserved.
  * This software is not subject to any license of Silicon Graphics 
  * Inc. or Purdue University.
  *
@@ -21,14 +24,8 @@
  * is preserved.
  */
 
-/*
- * A modify rdn request looks like this:
- *	ModifyRDNRequest ::= SEQUENCE {
- *		entry		DistinguishedName,
- *		newrdn		RelativeDistinguishedName,
- *		deleteoldrdn	BOOLEAN
- *		newSuperior	[0] DistinguishedName	[v3 only]
- *	}
+/* ACKNOWLEDGEMENTS:
+ * 	Juan C. Gomez
  */
 
 #include "portable.h"
@@ -41,6 +38,17 @@
 
 #include "ldap-int.h"
 
+/*
+ * A modify rdn request looks like this:
+ *	ModifyRDNRequest ::= SEQUENCE {
+ *		entry		DistinguishedName,
+ *		newrdn		RelativeDistinguishedName,
+ *		deleteoldrdn	BOOLEAN
+ *		newSuperior	[0] DistinguishedName	[v3 only]
+ *	}
+ */
+
+
 /*
  * ldap_rename - initiate an ldap extended modifyDN operation.
  *
@@ -67,6 +75,7 @@ ldap_rename(
 {
 	BerElement	*ber;
 	int rc;
+	ber_int_t id;
 
 	Debug( LDAP_DEBUG_TRACE, "ldap_rename\n", 0, 0, 0 );
 
@@ -79,6 +88,7 @@ ldap_rename(
 		return( LDAP_NO_MEMORY );
 	}
 
+	LDAP_NEXT_MSGID( ld, id );
 	if( newSuperior != NULL ) {
 		/* must be version 3 (or greater) */
 		if ( ld->ld_version < LDAP_VERSION3 ) {
@@ -86,15 +96,14 @@ ldap_rename(
 			ber_free( ber, 1 );
 			return( ld->ld_errno );
 		}
-
 		rc = ber_printf( ber, "{it{ssbtsN}", /* '}' */ 
-			++ld->ld_msgid, LDAP_REQ_MODDN,
+			id, LDAP_REQ_MODDN,
 			dn, newrdn, (ber_int_t) deleteoldrdn,
 			LDAP_TAG_NEWSUPERIOR, newSuperior );
 
 	} else {
 		rc = ber_printf( ber, "{it{ssbN}", /* '}' */ 
-			++ld->ld_msgid, LDAP_REQ_MODDN,
+			id, LDAP_REQ_MODDN,
 			dn, newrdn, (ber_int_t) deleteoldrdn );
 	}
 
@@ -118,7 +127,7 @@ ldap_rename(
 	}
 
 	/* send the message */
-	*msgidp = ldap_send_initial_request( ld, LDAP_REQ_MODRDN, dn, ber );
+	*msgidp = ldap_send_initial_request( ld, LDAP_REQ_MODRDN, dn, ber, id );
 	
 	if( *msgidp < 0 ) {
 		return( ld->ld_errno );
@@ -208,9 +217,9 @@ ldap_rename_s(
 		return rc;
 	}
 
-	rc = ldap_result( ld, msgid, 1, NULL, &res );
+	rc = ldap_result( ld, msgid, LDAP_MSG_ALL, NULL, &res );
 
-	if( rc == -1 ) {
+	if( rc == -1 || !res ) {
 		return ld->ld_errno;
 	}
 
