@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/libraries/liblutil/passfile.c,v 1.4.2.3 2005/01/20 17:01:04 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/libraries/liblutil/passfile.c,v 1.8.2.2 2007/08/31 23:13:57 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2007 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -44,7 +44,7 @@ lutil_get_filed_password(
 	}
 
 	passwd->bv_val = NULL;
-	passwd->bv_len = 4196;
+	passwd->bv_len = 4096;
 
 #ifdef HAVE_FSTAT
 	{
@@ -56,7 +56,8 @@ lutil_get_filed_password(
 					filename );
 			}
 
-			passwd->bv_len = sb.st_size;
+			if ( sb.st_size )
+				passwd->bv_len = sb.st_size;
 		}
 	}
 #endif /* HAVE_FSTAT */
