@@ -1,7 +1,16 @@
-/* $OpenLDAP: pkg/ldap/libraries/liblutil/sockpair.c,v 1.12.2.1 2003/03/03 17:10:06 kurt Exp $ */
-/*
- * Copyright 1998-2003 The OpenLDAP Foundation, All Rights Reserved.
- * COPYING RESTRICTIONS APPLY, see COPYRIGHT file
+/* $OpenLDAP: pkg/ldap/libraries/liblutil/sockpair.c,v 1.17.2.2 2007/08/31 23:13:57 quanah Exp $ */
+/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
+ *
+ * Copyright 1998-2007 The OpenLDAP Foundation.
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted only as authorized by the OpenLDAP
+ * Public License.
+ *
+ * A copy of this license is available in the file LICENSE in the
+ * top-level directory of the distribution or, alternatively, at
+ * <http://www.OpenLDAP.org/license.html>.
  */
 
 #include "portable.h"
@@ -26,7 +35,8 @@ int lutil_pair( ber_socket_t sds[2] )
 	return pipe( sds );
 #else
 	struct sockaddr_in si;
-	int rc, len = sizeof(si);
+	int rc;
+	ber_socklen_t len = sizeof(si);
 	ber_socket_t sd;
 
 	sd = socket( AF_INET, SOCK_DGRAM, 0 );
