@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/add.c,v 1.21.2.7 2005/01/20 22:43:28 ando Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/add.c,v 1.51.2.5 2007/08/31 23:14:03 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2005 The OpenLDAP Foundation.
+ * Copyright 1999-2007 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * Portions Copyright 1999-2003 Howard Chu.
  * All rights reserved.
@@ -35,74 +35,72 @@
 int
 meta_back_add( Operation *op, SlapReply *rs )
 {
-	struct metainfo *li = ( struct metainfo * )op->o_bd->be_private;
-	struct metaconn *lc;
-	int i, candidate = -1;
-	Attribute *a;
-	LDAPMod **attrs;
-	struct berval mdn = BER_BVNULL, mapped;
-	dncookie dc;
+	metainfo_t	*mi = ( metainfo_t * )op->o_bd->be_private;
+	metatarget_t	*mt;
+	metaconn_t	*mc;
+	int		i, candidate = -1;
+	int		isupdate;
+	Attribute	*a;
+	LDAPMod		**attrs;
+	struct berval	mdn = BER_BVNULL, mapped;
+	dncookie	dc;
+	int		msgid;
+	int		do_retry = 1;
+	LDAPControl	**ctrls = NULL;
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( BACK_META, ENTRY, "meta_back_add: %s\n",
-			op->o_req_dn.bv_val, 0, 0 );
-#else /* !NEW_LOGGING */
 	Debug(LDAP_DEBUG_ARGS, "==> meta_back_add: %s\n",
 			op->o_req_dn.bv_val, 0, 0 );
-#endif /* !NEW_LOGGING */
 
 	/*
 	 * get the current connection
 	 */
-	lc = meta_back_getconn( op, rs, META_OP_REQUIRE_SINGLE,
-			&op->o_req_ndn, &candidate );
-	if ( !lc ) {
-		send_ldap_result( op, rs );
+	mc = meta_back_getconn( op, rs, &candidate, LDAP_BACK_SENDERR );
+	if ( !mc || !meta_back_dobind( op, rs, mc, LDAP_BACK_SENDERR ) ) {
+		return rs->sr_err;
 	}
 
-	if ( !meta_back_dobind( lc, op )
-			|| !meta_back_is_valid( lc, candidate ) ) {
-		rs->sr_err = LDAP_OTHER;
- 		send_ldap_result( op, rs );
-		return -1;
-	}
+	assert( mc->mc_conns[ candidate ].msc_ld != NULL );
 
 	/*
 	 * Rewrite the add dn, if needed
 	 */
-	dc.rwmap = &li->targets[ candidate ]->rwmap;
+	mt = mi->mi_targets[ candidate ];
+	dc.target = mt;
 	dc.conn = op->o_conn;
 	dc.rs = rs;
 	dc.ctx = "addDN";
 
 	if ( ldap_back_dn_massage( &dc, &op->o_req_dn, &mdn ) ) {
 		send_ldap_result( op, rs );
-		return -1;
+		goto done;
 	}
 
-	/* Count number of attributes in entry */
-	for ( i = 1, a = op->oq_add.rs_e->e_attrs; a; i++, a = a->a_next );
+	/* Count number of attributes in entry ( +1 ) */
+	for ( i = 1, a = op->ora_e->e_attrs; a; i++, a = a->a_next );
 	
 	/* Create array of LDAPMods for ldap_add() */
 	attrs = ch_malloc( sizeof( LDAPMod * )*i );
 
-	for ( i = 0, a = op->oq_add.rs_e->e_attrs; a; a = a->a_next ) {
-		int	j, is_oc = 0;
+	dc.ctx = "addAttrDN";
+	isupdate = be_shadow_update( op );
+	for ( i = 0, a = op->ora_e->e_attrs; a; a = a->a_next ) {
+		int			j, is_oc = 0;
 
-		if ( a->a_desc->ad_type->sat_no_user_mod  ) {
+		if ( !isupdate && !get_relax( op ) && a->a_desc->ad_type->sat_no_user_mod  )
+		{
 			continue;
 		}
 
-		if ( a->a_desc == slap_schema.si_ad_objectClass
+		if ( a->a_desc == slap_schema.si_ad_objectClass 
 				|| a->a_desc == slap_schema.si_ad_structuralObjectClass )
 		{
 			is_oc = 1;
 			mapped = a->a_desc->ad_cname;
 
 		} else {
-			ldap_back_map( &li->targets[ candidate ]->rwmap.rwm_at,
+			ldap_back_map( &mt->mt_rwmap.rwm_at,
 					&a->a_desc->ad_cname, &mapped, BACKLDAP_MAP );
-			if ( mapped.bv_val == NULL || mapped.bv_val[0] == '\0' ) {
+			if ( BER_BVISNULL( &mapped ) || BER_BVISEMPTY( &mapped ) ) {
 				continue;
 			}
 		}
@@ -125,11 +123,11 @@ meta_back_add( Operation *op, SlapReply *rs )
 			for ( j = 0; !BER_BVISNULL( &a->a_vals[ j ] ); ) {
 				struct ldapmapping	*mapping;
 
-				ldap_back_mapping( &li->targets[ candidate ]->rwmap.rwm_oc,
+				ldap_back_mapping( &mt->mt_rwmap.rwm_oc,
 						&a->a_vals[ j ], &mapping, BACKLDAP_MAP );
 
 				if ( mapping == NULL ) {
-					if ( li->targets[ candidate ]->rwmap.rwm_oc.drop_missing ) {
+					if ( mt->mt_rwmap.rwm_oc.drop_missing ) {
 						continue;
 					}
 					attrs[ i ]->mod_bvalues[ j ] = &a->a_vals[ j ];
@@ -139,12 +137,6 @@ meta_back_add( Operation *op, SlapReply *rs )
 				}
 				j++;
 			}
-
-			if ( j == 0 ) {
-				ch_free( attrs[ i ]->mod_bvalues );
-				continue;
-			}
-
 			attrs[ i ]->mod_bvalues[ j ] = NULL;
 
 		} else {
@@ -153,16 +145,20 @@ meta_back_add( Operation *op, SlapReply *rs )
 			 * to allow their use in ACLs at the back-ldap
 			 * level.
 			 */
-			if ( a->a_desc->ad_type->sat_syntax == slap_schema.si_syn_distinguishedName ) {
+			if ( a->a_desc->ad_type->sat_syntax ==
+				slap_schema.si_syn_distinguishedName )
+			{
 				(void)ldap_dnattr_rewrite( &dc, a->a_vals );
 				if ( a->a_vals == NULL ) {
 					continue;
 				}
 			}
 
-			for ( j = 0; a->a_vals[ j ].bv_val; j++ );
-			attrs[ i ]->mod_bvalues = ch_malloc((j+1)*sizeof(struct berval *));
-			for ( j = 0; a->a_vals[ j ].bv_val; j++ ) {
+			for ( j = 0; !BER_BVISNULL( &a->a_vals[ j ] ); j++ )
+				;
+			
+			attrs[ i ]->mod_bvalues = ch_malloc( ( j + 1 ) * sizeof( struct berval * ) );
+			for ( j = 0; !BER_BVISNULL( &a->a_vals[ j ] ); j++ ) {
 				attrs[ i ]->mod_bvalues[ j ] = &a->a_vals[ j ];
 			}
 			attrs[ i ]->mod_bvalues[ j ] = NULL;
@@ -171,15 +167,45 @@ meta_back_add( Operation *op, SlapReply *rs )
 	}
 	attrs[ i ] = NULL;
 
-	ldap_add_s( lc->conns[ candidate ].ld, mdn.bv_val, attrs );
+retry:;
+	ctrls = op->o_ctrls;
+	if ( meta_back_controls_add( op, rs, mc, candidate, &ctrls ) != LDAP_SUCCESS )
+	{
+		send_ldap_result( op, rs );
+		goto cleanup;
+	}
+
+	rs->sr_err = ldap_add_ext( mc->mc_conns[ candidate ].msc_ld, mdn.bv_val,
+			      attrs, ctrls, NULL, &msgid );
+	rs->sr_err = meta_back_op_result( mc, op, rs, candidate, msgid,
+		mt->mt_timeout[ SLAP_OP_ADD ], LDAP_BACK_SENDRESULT );
+	if ( rs->sr_err == LDAP_UNAVAILABLE && do_retry ) {
+		do_retry = 0;
+		if ( meta_back_retry( op, rs, &mc, candidate, LDAP_BACK_SENDERR ) ) {
+			/* if the identity changed, there might be need to re-authz */
+			(void)ldap_back_controls_free( op, rs, &ctrls );
+			goto retry;
+		}
+	}
+
+cleanup:;
+	(void)ldap_back_controls_free( op, rs, &ctrls );
+
 	for ( --i; i >= 0; --i ) {
 		free( attrs[ i ]->mod_bvalues );
 		free( attrs[ i ] );
 	}
 	free( attrs );
-	if ( mdn.bv_val != op->oq_add.rs_e->e_dn ) {
+	if ( mdn.bv_val != op->ora_e->e_dn ) {
 		free( mdn.bv_val );
+		BER_BVZERO( &mdn );
 	}
-	return meta_back_op_result( lc, op, rs );
+
+done:;
+	if ( mc ) {
+		meta_back_release_conn( mi, mc );
+	}
+
+	return rs->sr_err;
 }
 
