@@ -1,8 +1,8 @@
 /* ctxcsn.c -- Context CSN Management Routines */
-/* $OpenLDAP: pkg/ldap/servers/slapd/ctxcsn.c,v 1.9.2.11 2005/01/20 17:01:06 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/ctxcsn.c,v 1.40.2.4 2007/08/31 23:13:58 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2003-2005 The OpenLDAP Foundation.
+ * Copyright 2003-2007 The OpenLDAP Foundation.
  * Portions Copyright 2003 IBM Corporation.
  * All rights reserved.
  *
@@ -22,22 +22,25 @@
 #include <ac/string.h>
 #include <ac/socket.h>
 
-#include "ldap_pvt.h"
 #include "lutil.h"
 #include "slap.h"
 #include "lutil_ldap.h"
 
 const struct berval slap_ldapsync_bv = BER_BVC("ldapsync");
 const struct berval slap_ldapsync_cn_bv = BER_BVC("cn=ldapsync");
+int slap_serverID;
 
 void
-slap_get_commit_csn( Operation *op, struct berval *csn )
+slap_get_commit_csn(
+	Operation *op,
+	struct berval *maxcsn
+)
 {
 	struct slap_csn_entry *csne, *committed_csne = NULL;
-	int i = 0;
 
-	csn->bv_val = NULL;
-	csn->bv_len = 0;
+	if ( maxcsn ) {
+		BER_BVZERO( maxcsn );
+	}
 
 	ldap_pvt_thread_mutex_lock( op->o_bd->be_pcl_mutexp );
 
@@ -53,7 +56,7 @@ slap_get_commit_csn( Operation *op, struct berval *csn )
 		if ( csne->ce_state == SLAP_CSN_PENDING ) break;
 	}
 
-	if ( committed_csne ) ber_dupbv_x( csn, committed_csne->ce_csn, op->o_tmpmemctx );
+	if ( committed_csne && maxcsn ) *maxcsn = committed_csne->ce_csn;
 	ldap_pvt_thread_mutex_unlock( op->o_bd->be_pcl_mutexp );
 }
 
@@ -82,14 +85,24 @@ slap_graduate_commit_csn( Operation *op )
 	if ( op == NULL ) return;
 	if ( op->o_bd == NULL ) return;
 
+#if 0
+	/* it is NULL when we get here from the frontendDB;
+	 * alternate fix: initialize frontendDB like all other backends */
+	assert( op->o_bd->be_pcl_mutexp != NULL );
+#endif
+	
+	if ( op->o_bd->be_pcl_mutexp == NULL ) return;
+
 	ldap_pvt_thread_mutex_lock( op->o_bd->be_pcl_mutexp );
 
 	LDAP_TAILQ_FOREACH( csne, op->o_bd->be_pending_csn_list, ce_csn_link ) {
 		if ( csne->ce_opid == op->o_opid && csne->ce_connid == op->o_connid ) {
 			LDAP_TAILQ_REMOVE( op->o_bd->be_pending_csn_list,
 				csne, ce_csn_link );
-			ch_free( csne->ce_csn->bv_val );
-			ch_free( csne->ce_csn );
+			if ( op->o_csn.bv_val == csne->ce_csn.bv_val ) {
+				BER_BVZERO( &op->o_csn );
+			}
+			ch_free( csne->ce_csn.bv_val );
 			ch_free( csne );
 			break;
 		}
@@ -104,7 +117,7 @@ static struct berval ocbva[] = {
 	BER_BVC("top"),
 	BER_BVC("subentry"),
 	BER_BVC("syncProviderSubentry"),
-	{0,NULL}
+	BER_BVNULL
 };
 
 Entry *
@@ -113,11 +126,10 @@ slap_create_context_csn_entry(
 	struct berval *context_csn )
 {
 	Entry* e;
-	int rc;
 
 	struct berval bv;
 
-	e = (Entry *) ch_calloc( 1, sizeof( Entry ));
+	e = entry_alloc();
 
 	attr_merge( e, slap_schema.si_ad_objectClass,
 		ocbva, NULL );
@@ -131,8 +143,7 @@ slap_create_context_csn_entry(
 			context_csn, NULL );
 	}
 
-	bv.bv_val = "{}";
-	bv.bv_len = sizeof("{}")-1;
+	BER_BVSTR( &bv, "{}" );
 	attr_merge_one( e, slap_schema.si_ad_subtreeSpecification, &bv, NULL );
 
 	build_new_dn( &e->e_name, &be->be_nsuffix[0],
@@ -142,34 +153,42 @@ slap_create_context_csn_entry(
 	return e;
 }
 
-int
-slap_get_csn(
+void
+slap_queue_csn(
 	Operation *op,
-	char *csnbuf,
-	int	len,
-	struct berval *csn,
-	int manage_ctxcsn )
+	struct berval *csn )
 {
 	struct slap_csn_entry *pending;
 
+	pending = (struct slap_csn_entry *) ch_calloc( 1,
+			sizeof( struct slap_csn_entry ));
+	ldap_pvt_thread_mutex_lock( op->o_bd->be_pcl_mutexp );
+
+	ber_dupbv( &pending->ce_csn, csn );
+	ber_bvreplace_x( &op->o_csn, &pending->ce_csn, op->o_tmpmemctx );
+	pending->ce_connid = op->o_connid;
+	pending->ce_opid = op->o_opid;
+	pending->ce_state = SLAP_CSN_PENDING;
+	LDAP_TAILQ_INSERT_TAIL( op->o_bd->be_pending_csn_list,
+		pending, ce_csn_link );
+	ldap_pvt_thread_mutex_unlock( op->o_bd->be_pcl_mutexp );
+}
+
+int
+slap_get_csn(
+	Operation *op,
+	struct berval *csn,
+	int manage_ctxcsn )
+{
 	if ( csn == NULL ) return LDAP_OTHER;
 
-	csn->bv_len = lutil_csnstr( csnbuf, len, 0, 0 );
-	csn->bv_val = csnbuf;
+	/* gmtime doesn't always need a mutex, but lutil_csnstr does */
+	ldap_pvt_thread_mutex_lock( &gmtime_mutex );
+	csn->bv_len = lutil_csnstr( csn->bv_val, csn->bv_len, slap_serverID, 0 );
+	ldap_pvt_thread_mutex_unlock( &gmtime_mutex );
 
-	if ( manage_ctxcsn ) {
-		pending = (struct slap_csn_entry *) ch_calloc( 1,
-			sizeof( struct slap_csn_entry ));
-		ldap_pvt_thread_mutex_lock( op->o_bd->be_pcl_mutexp );
-		ber_dupbv( &op->o_sync_csn, csn );
-		pending->ce_csn = ber_dupbv( NULL, csn );
-		pending->ce_connid = op->o_connid;
-		pending->ce_opid = op->o_opid;
-		pending->ce_state = SLAP_CSN_PENDING;
-		LDAP_TAILQ_INSERT_TAIL( op->o_bd->be_pending_csn_list,
-			pending, ce_csn_link );
-		ldap_pvt_thread_mutex_unlock( op->o_bd->be_pcl_mutexp );
-	}
+	if ( manage_ctxcsn )
+		slap_queue_csn( op, csn );
 
 	return LDAP_SUCCESS;
 }
